unit UMainForm;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StrUtils, StdCtrls, ExtCtrls,
  UPiece,
  UCube,
  UPuzzle,
  Math;

type

  TViewType = (vtCompact = 0, vtSemiExploded = 1, vtExploded = 2);

  TMainForm = class(TForm)
    mPiecesAndViews: TMemo;
    btResolve: TButton;
    mPuzzle: TMemo;
    mCurrentTest: TMemo;
    lTestCount: TLabel;
    lStats: TLabel;
    lSolCount: TLabel;
    Image1: TImage;
    Image2: TImage;
    cbDrawSolution: TCheckBox;
    rgDisplayType: TRadioGroup;
    btLoadPuzzle: TButton;
    ePuzzleFileName: TEdit;
    mTime: TMemo;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    btStop: TButton;
    cbWriteSolutions: TCheckBox;

    procedure FormCreate(Sender: TObject);
    procedure btResolveClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure rgDisplayTypeClick(Sender: TObject);
    procedure btLoadPuzzleClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btStopClick(Sender: TObject);
  private
    function LoadPuzzle: boolean;
    procedure LoadPieceSet;

    function CheckPosition(_MinCube: TCube; _PuzzleCoords, _PieceVueCoords: TCubeList): boolean;
    function ChangeUsedState(_MinCube: TCube; _PuzzleCoords, _PieceVueCoords: TCubeList; _PieceIx: integer = -1): boolean;

    procedure DisenabledCtrls(_enabled: boolean = true);
    procedure Resolve(_Deep: integer);

    procedure SaveSolution;
    procedure AddTimeLine(str: string);
    procedure RefreshStats;

    //Partie graphique
    procedure DrawPuzzle;
    procedure DrawSolution;
    procedure CalculateBmpSize(var _XMin, _XMax, _YMin, _YMax: integer);
    procedure DrawACube(const _Cube: TCube; var _Bmp: TBitmap; const _Color: TColor);

  public
    FPuzzle: TPuzzle;
    FPieces: TPieceSet;
    FTotalCubeOfPieces: integer;

    FSolCount: integer;
    FTest: integer;
    FSolutionFile: TextFile;

    FProgBeginAt: TDateTime;
    FSolveBeginAt: TDateTime;
    FPause: boolean;
    FHasToExit: boolean;

    FViewType: TViewType;
    FPuzzleBmp: TBitmap;
    FBasePoint: TPoint;
  end;

const
  kPuzzleDirectory = 'PuzzleFiles';
  kDefaultPuzzle = 'SOMA.csv';
  kCurrentLine = '- Cube %d %s, Piece: %s, Vue %d';
  kSolutionFile = 'Solution.txt';

var
  MainForm: TMainForm;
  FApplicationPath: string;

implementation

{$R *.dfm}

//----------------------------------------------------------------------------------------------------------------------
procedure TMainForm.FormCreate(Sender: TObject);
begin
  FProgBeginAt:= Now;
  
  FApplicationPath:= IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName));

  if ParamCount = 1 then
    ePuzzleFileName.Text:= ParamStr(1)
  else
    ePuzzleFileName.Text:=
      IncludeTrailingPathDelimiter(FApplicationPath + kPuzzleDirectory) + kDefaultPuzzle;

  mTime           .Clear;
  mPiecesAndViews .Clear;
  mPuzzle         .Clear;
  mCurrentTest    .Clear;

  FPuzzle:= nil;
  FPieces:= nil;

  rgDisplayType.ItemIndex:= 2;
end;


//------------------------------------------------------------------------------------------------------
procedure TMainForm.FormActivate(Sender: TObject);
begin
  LoadPieceSet;
  
  LoadPuzzle;

  btResolveClick(nil);
end;

//----------------------------------------------------------------------------------------------------------------------
//  Permet l'arrt de l'algorithme si demande de fermeture
procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  FHasToExit:= true;
end;

//----------------------------------------------------------------------------------------------------------------------
procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FPuzzle.Free;
  FPieces.Free;
end;

//------------------------------------------------------------------------------------------------------
procedure TMainForm.LoadPieceSet;
var
  aPiece: TPiece;
  i, j: integer;
begin
  if (FPieces <> nil) then
    FreeAndNil(FPieces);
  FPieces:= TPieceSet.Create;

  AddTimeLine('Chargement des pices');

  //Piece P
  aPiece:= TPiece.Create('P', clRed);
  aPiece.BaseCubeList.Add(TCube.Create(0,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(1,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(0,1,0));
  aPiece.BaseCubeList.Add(TCube.Create(0,0,1));
  FPieces.Add(aPiece);

  //Piece A
  aPiece:= TPiece.Create('A', clTeal);
  aPiece.BaseCubeList.Add(TCube.Create(0,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(0,1,0));
  aPiece.BaseCubeList.Add(TCube.Create(0,0,1));
  aPiece.BaseCubeList.Add(TCube.Create(1,0,1));
  FPieces.Add(aPiece);

  //Piece B
  aPiece:= TPiece.Create('B', clGreen);
  aPiece.BaseCubeList.Add(TCube.Create(0,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(1,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(0,0,1));
  aPiece.BaseCubeList.Add(TCube.Create(0,1,1));
  FPieces.Add(aPiece);

  //Piece V
  aPiece:= TPiece.Create('V', clSilver);
  aPiece.BaseCubeList.Add(TCube.Create(0,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(1,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(0,1,0));
  FPieces.Add(aPiece);

  //Piece L
  aPiece:= TPiece.Create('L', clBlue);
  aPiece.BaseCubeList.Add(TCube.Create(0,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(1,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(0,1,0));
  aPiece.BaseCubeList.Add(TCube.Create(0,2,0));
  FPieces.Add(aPiece);

  //Piece T
  aPiece:= TPiece.Create('T', clYellow);
  aPiece.BaseCubeList.Add(TCube.Create(0,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(1,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(2,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(1,1,0));
  FPieces.Add(aPiece);

  //Piece Z
  aPiece:= TPiece.Create('Z', clLime);
  aPiece.BaseCubeList.Add(TCube.Create(0,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(1,0,0));
  aPiece.BaseCubeList.Add(TCube.Create(1,1,0));
  aPiece.BaseCubeList.Add(TCube.Create(2,1,0));
  FPieces.Add(aPiece);

  AddTimeLine('Pices charges');

  //Calcul des vues de chaque pice
  for i:= 0 to FPieces.Count - 1 do
    with TPiece(FPieces.Items[i]) do
    begin
      mPiecesAndViews.Lines.Add(GetBaseCubeString);
      ComputeAngleViews;
      for j:= 0 to AngleViewList.Count - 1 do
        mPiecesAndViews.Lines.Add(GetAngleViewString(j));
    end;

  AddTimeLine('Angles de vue calculs');

end;

//------------------------------------------------------------------------------------------------------
function TMainForm.LoadPuzzle: boolean;
var
  i: integer;
begin
  AddTimeLine('Chargement du puzzle');

  if (FPuzzle <> nil) then
    FreeAndNil(FPuzzle);

  //Chargement et affichage du puzzle
  FPuzzle:= TPuzzle.Create;

  if not ( FPuzzle.LoadFromFile(ePuzzleFileName.Text) ) then
  begin
    AddTimeLine('Erreur : fichier puzzle incorrect');
    ShowMessage('Erreur : fichier puzzle incorrect');
    FreeAndNil(FPuzzle);
    Exit;
  end;

  DrawPuzzle;
  mPuzzle.Clear;
  for i:= 0 to FPuzzle.Coords.Count - 1 do
    mPuzzle.Lines.Add(TCube(FPuzzle.Coords[i]).GetString);

  AddTimeLine('Puzzle charg');


  if ( FPuzzle.Coords.Count <> FPieces.CubesCount ) then
  begin
    ShowMessage('Ce puzzle ne peut-tre rsolu, il ne contient pas le bon nombre de cube');
    FreeAndNil(FPuzzle);
  end;

end;

//----------------------------------------------------------------------------------------------------------------------
// Vrifie que les cubes de la pice-vue peuvent se placer  la position en cours du puzzle
function TMainForm.CheckPosition(_MinCube: TCube; _PuzzleCoords, _PieceVueCoords: TCubeList): boolean;
var
  i: integer;
  acube: TCube;
begin
  result:= true;

  //Pour chaque cube de la pice dans la vue en cours
  for i:= 0 to _PieceVueCoords.Count - 1 do
  begin

    acube:= TCube.AddCoords(_MinCube, TCube(_PieceVueCoords[i]));

    //Ce cube appartient-il au puzzle et est-il inutilis
    if ( _PuzzleCoords.IsInListAndNotUsed(acube) = -1) then
    begin
      result:= false;
      acube.Free;
      break;
    end;
    
    acube.Free;
  end;

end;

//----------------------------------------------------------------------------------------------------------------------
//  Permet de marquer les cubes du puzzle pris par une pice
function TMainForm.ChangeUsedState(_MinCube: TCube;
 _PuzzleCoords, _PieceVueCoords: TCubeList; _PieceIx: integer = -1): boolean;
var
  i, res: integer;
  acube: TCube;
begin
  result:= true;

  for i:= 0 to _PieceVueCoords.Count - 1 do
  begin
    acube:= TCube.AddCoords(_MinCube, TCube(_PieceVueCoords[i]));
    res:= _PuzzleCoords.IsInList(acube);
    TCube(_PuzzleCoords[res]).UsedByPiece:= _PieceIx;
    acube.Free;
  end;

end;


//----------------------------------------------------------------------------------------------------------------------
procedure TMainForm.btResolveClick(Sender: TObject);
var
  i: integer;
  str: string;
begin
  if (FPuzzle = nil) then
  begin
    ShowMessage('Veuillez d''abord charger un puzzle valide');
    Exit
  end;

  DisenabledCtrls(false);

  //Initilisation
  mCurrentTest.Clear;
  for i:= 0 to FPieces.Count - 1 do
    mCurrentTest.lines.Add('');

  FSolCount:= 0;
  FTest:= 0;
  FSolveBeginAt:= Now;
  FHasToExit:= false;

  //Initialisation fichier de sortie
  if cbWriteSolutions.Checked then
  begin
    AssignFile(FSolutionFile, FApplicationPath + kSolutionFile);
    Rewrite(FSolutionFile);
    Writeln(FSolutionFile, ePuzzleFileName.Text);
    Writeln(FSolutionFile, mPiecesAndViews.Lines.Text);
  end;

  AddTimeLine('Lancement rsolution du puzzle : ' + ePuzzleFileName.Text);

  //Appel de l'algo de rsolution
  Resolve(0);

  if FHasToExit then
    AddTimeLine('Rsolution stopp par l''utilisateur')
  else
    AddTimeLine('Rsolution finie');

  AddTimeLine(Format('%d solution(s) trouve(s)', [FSolCount]));

  if cbWriteSolutions.Checked then
    CloseFile(FSolutionFile);

  // Fin de l'algorithme, on affiche le nombre exacte de test et le temps
  RefreshStats;

  // L'algorithme est fini et aucune solution n'a t trouv
  if ( FSolCount = 0 ) then
    ShowMessage('Le puzzle n''a pas de solution !!');

  DisenabledCtrls;
end;

//----------------------------------------------------------------------------------------------------------------------
procedure TMainForm.btStopClick(Sender: TObject);
begin
  FHasToExit:= true;
  DisenabledCtrls;
end;

//----------------------------------------------------------------------------------------------------------------------
procedure TMainForm.DisenabledCtrls(_enabled: boolean = true);
begin
  btResolve.Enabled:= _enabled;
  btLoadPuzzle.Enabled:= _enabled;
  cbWriteSolutions.Enabled:= _enabled;
end;

//----------------------------------------------------------------------------------------------------------------------
procedure TMainForm.RefreshStats;
var
  h, m, s, n: WORD;
begin
  lTestCount.Caption:= Format('Tests : %d', [FTest]);

  DecodeTime(Now - FSolveBeginAt, h, m, s, n);
  lStats.Caption:= FormatDateTime('hh:nn:ss:zzz', Now - FSolveBeginAt)
     + Format(' / %.0f tests/s', [ Ftest/(60*m + s + 1) ]);
  // +1, vite la division par 0 au dbut quand m et s valent 0
end;

//----------------------------------------------------------------------------------------------------------------------
procedure TMainForm.Resolve(_Deep: integer);
var
  piece, vue, posmin: integer;
  isp: boolean;
  cMin: TCube;
  vueCoords: TCubeList;
begin
  //Possibilite de stopper l'algorithme
  if FHasToExit then exit;

  //on rcupre le cube minimun  rsoudre
  posmin:= FPuzzle.Coords.NotUsedMinCube;
  cMin:= TCube(FPuzzle.Coords[posmin]);

  for piece:= 0 to FPieces.Count - 1 do             //Pour chaque pice

    if not TPiece(FPieces[piece]).Used then         //Si la pice est libre
      with TPiece(FPieces[piece]) do
        for vue:= 0 to AngleViewList.Count - 1 do   //On vrifie la possibilit de chaque vue
        begin
          //Evitons de bloquer le prog
          Application.ProcessMessages;

          inc(Ftest);
          vueCoords:= TCubeList(AngleViewList[vue]);

          //Test de la possibilit de la vue
          isp:= CheckPosition(cMin, FPuzzle.Coords, vueCoords);

          //La vue est possible
          if isp then
          begin
            mCurrentTest.Lines[_Deep]:=
              Format(kCurrentLine, [posmin, cMin.GetString, Name, vue]);

            //On marque les cubes du puzzle comme tant utiliss par la pice en cours
            ChangeUsedState(cMin, FPuzzle.Coords, vueCoords, piece);

            //On marque la pice comme tant utilise aussi
            Used:= true;

            //Il y a encore des pices  placer, on appel resolve pour rsoudre le sous-puzzle restant
            //O sous-puzzle est le puzzle moins les cubes et les pices dj utiliss
            if (_Deep < FPieces.Count - 1) then
                Resolve(_Deep + 1)

            //La vue est possible et il n'y a plus de pice  placer, on a donc une solution
            else if ( _Deep = FPieces.Count - 1 ) then
            begin
              inc(FSolCount);

              //Affichage du temps pour trouver une 1ere solution
              if (FSolCount = 1) then
                AddTimeLine('Premire solution aprs : '
                 + FormatDateTime('hh:nn:ss:zzz', Now - FSolveBeginAt));

              lSolCount.Caption:= Format('Solution(s): %d' , [FSolCount]);

              //Ecriture de la solution dans le fichier de sortie
              if ( cbWriteSolutions.Checked ) then
                SaveSolution;

              //Affichage de la solution
              if ( cbDrawSolution.Checked ) then
                DrawSolution;

            end;

            //Avant de sortir on marque les cubes et la pice comme tant inutiliss
            ChangeUsedState(cMin, FPuzzle.Coords, vueCoords);
            Used:= false;

          end;

          //On rafraichit les stats tout les 1000 tests
          if ( FTest mod 1000 = 0) then
            RefreshStats;

        end;

end;

//----------------------------------------------------------------------------------------------------------------------
procedure TMainForm.AddTimeLine(str: string);
begin
  mTime.Lines.Add( FormatDateTime('hh:nn:ss:zzz', Now - FProgBeginAt) + ' - ' + str );
end;

//-----------------------------------------------------------------------------------------------------------
procedure TMainForm.SaveSolution;
var
  i: integer;
begin
  WriteLn(FSolutionFile, 'Solution ' + inttostr(FSolCount));
  writeLn(FSolutionFile, mCurrentTest.Lines.Text);
  for i:= 0 to FPuzzle.Coords.Count - 1 do
    with TCube(FPuzzle.Coords[i]) do
      Writeln(FSolutionFile, Format('%d,%d,%d,%s', [X, Y, Z, TPiece(FPieces[UsedByPiece]).Name]));

  WriteLn(FSolutionFile, '');
end;

//----------------------------------------------------------------------------------------------------------------------
// But: permet de calculer la taille du bitmap en fonction du puzzle et de la vue
procedure TMainForm.CalculateBmpSize(var _XMin, _XMax, _YMin, _YMax: integer);
var
  di, dj: double;
  i, dx, dy: integer;
begin
  //en vue explose di est diffrent de dj pour permettre de voir tout les cubes
  //une vue purement isomtrique di=dj mme clate ne le permet pas
  case FViewType of
    vtCompact     : begin di:= 1;   dj:= 1;   end;
    vtSemiExploded: begin di:= 1.5; dj:= 1.5; end;
    vtExploded    : begin di:= 2.2; dj:= 2.5; end;
  end;

  with TCube(FPuzzle.Coords[0]) do
  begin
    dx:= Round(10*dj*X + 10*di*Z);
    dy:= Round(-10*dj*X + 10*di*(9 - Y) + 10*di*Z);
    //on inverse en Y, on veut Y vers le haut, mais les pixels sont ordonnes vers le bas
  end;

  _XMin:= dx;
  _XMax:= dx;
  _YMin:= dy;
  _YMax:= dy;

  for i:= 1 to FPuzzle.Coords.Count - 1 do
    with TCube(FPuzzle.Coords[i]) do
    begin

      dx:= Round(10*dj*X + 10*di*Z);
      dy:= Round(-10*dj*X + 10*di*(9 - Y) + 10*di*Z);
      //on inverse en Y, on veut Y vers le haut, mais les pixels sont ordonnes vers le bas

      _XMin:= Min(_XMin, dx);
      _XMax:= Max(_XMax, dx);
      _YMin:= Min(_YMin, dy);
      _YMax:= Max(_YMax, dy);

    end;

end;

//----------------------------------------------------------------------------------------------------------------------
//  Dessine un cube dans le bitmap en fonction de ces coordonnes et de la vue
procedure TMainForm.DrawACube(const _Cube: TCube; var _Bmp: TBitmap; const _Color: TColor);
var
  di, dj: double;
  x, y: integer;
begin
  case FViewType of
    vtCompact     : begin di:= 1;   dj:= 1;   end;
    vtSemiExploded: begin di:= 1.5; dj:= 1.5; end;
    vtExploded    : begin di:= 2.2; dj:= 2.5; end;
  end;

  //Calcul de la position de la base du cube
  x:= FBasePoint.X + Round(10*dj*_Cube.X + 10*di*_Cube.Z);
  y:= FBasePoint.Y + Round(-10*dj*_Cube.X + 10*di*(9 - _Cube.Y) + 10*di*_Cube.Z);
  //on inverse en Y, on veut Y vers le haut, mais les pixels sont ordonnes vers le bas

  with _Bmp.Canvas do
  begin
    Pen.Width:= 1;
    Pen.Style:= psSolid;
    Brush.Style:= bsSolid;
    Brush.Color:= _Color;

    //Contour du cube
    Polygon([ Point(x     , y + 10),
              Point(x + 10, y     ),
              Point(x + 20, y + 10),
              Point(x + 20, y + 20),
              Point(x + 10, y + 30),
              Point(x     , y + 20)]);

    //Ligne intrieur
    MoveTo(x     , y + 10);
    LineTo(x + 10, y + 20);
    LineTo(x + 20, y + 10);

    MoveTo(x + 10, y + 20);
    LineTo(x + 10, y + 30);
  end;


end;

//-----------------------------------------------------------------------------------------------------------
procedure TMainForm.DrawPuzzle;
var
  j, xmin, xmax, ymin, ymax: integer;
begin
  if ( FPuzzleBmp <> nil ) then
    FreeAndNil(FPuzzleBmp);

  FPuzzleBmp:= TBitmap.Create;

  CalculateBmpSize(xmin, xmax, ymin, ymax);

  FPuzzleBmp.Width  := (xmax - xmin + 21);
  FPuzzleBmp.Height := (ymax - ymin + 30);
  FBasePoint:= Point(-xmin, -ymin);

  FPuzzle.Coords.SortForDrawing;

  for j:= FPuzzle.Coords.Count - 1 downto 0 do
    DrawACube(TCube(FPuzzle.Coords[j]), FPuzzleBmp, clWhite);

  Image1.Picture.Assign(FPuzzleBmp);

  FPuzzle.Coords.SortByMin;
end;


//-----------------------------------------------------------------------------------------------------------
procedure TMainForm.DrawSolution;
var
  i: integer;
  SolutionBmp: TBitmap;
begin
  SolutionBmp:= TBitmap.Create;
  SolutionBmp.Width:= FPuzzleBmp.Width;
  SolutionBmp.Height:= FPuzzleBmp.Height;

  FPuzzle.Coords.SortForDrawing;

  Image2.Picture:= nil;

  for i:= FPuzzle.Coords.Count - 1 downto 0 do
  begin
    DrawACube(TCube(FPuzzle.Coords[i]), SolutionBmp,
      TPiece(FPieces[TCube(FPuzzle.Coords[i]).UsedByPiece] ).Color);
  end;

  Image2.Picture.Assign(SolutionBmp);

  FPuzzle.Coords.SortByMin;

  SolutionBmp.Free;
end;


//----------------------------------------------------------------------------------------------------------------------
procedure TMainForm.rgDisplayTypeClick(Sender: TObject);
begin
  FViewType:= TViewType(rgDisplayType.ItemIndex);

  if (FPuzzle <> nil) then
    DrawPuzzle;
end;

//----------------------------------------------------------------------------------------------------------------------
procedure TMainForm.btLoadPuzzleClick(Sender: TObject);
begin

  with TOpenDialog.Create(Self) do
  begin

    InitialDir:= ExtractFilePath(Application.ExeName) + kPuzzleDirectory;
    Filter:= 'Puzzle File (*.txt, *.csv) |*.txt; *csv';

    if Execute then
    begin
      ePuzzleFileName.Text:= FileName;
      LoadPuzzle;
    end;

  end;

end;

//----------------------------------------------------------------------------------------------------------------------
end.
