unit UPiece;

interface

uses
  SysUtils,
  Classes,
  UCube,
  IniFiles,
  Graphics;


type

(*
  Pour le problme pos,

  E = {0, 1, .., 9} = les entiers de 0  9;

  Nous considrons une pice comme un ensemble (de 3  4) de coordonnes dans l'espace F = E*E*E
  Dans F, la pice  un certain nombre de position

  Si l'on calcule 4 rotations en X (5 ramenant  la position de dpart), par 4
  rotations en Y et par 4 en Z, on obtient 4*4*4 = 64 angles de vues.

  On peut montrer qu'il n'y a en fait que 48 angle de vues (par exemple
   2 rotations sur X, sur Y et sur Z ramne la pice  sa position d'origine)

  L'ensemble des rotations donnent pour un cube la possibilit de prendre des
  coordonnes dans l'ensemble (x, -x, y, -y, z, -z) avec seulement une en x, une en y, une en z
  soit pour la premire coordonnes : (x, -x, y, -y, z, -z) = 6
  pour la seconde (en fonction de la 1ere): (y, -y, z, -z) ou (x, -x, z, -z) ou (x, -x, y, -y) = 4
  pour la troisime : (x, -x) ou (y, -y) ou (z, -z) = 2
  soit 6*4*2 = 48
  N'ayant pas trouver d'algorithme (dans le temps) pour dcrire ces 48 rotations,
  je calcule l'ensemble des 64 bien que certaines soient similaires. Et limine au fur
  et  mesure les invariances (function AngleViewIsInList) ce qui est de toute faon,
  mme en en calculant que 48, ncessaire car les symtries des pices implique
  des invariances.

  Les rotaions sont notes
  
  Fx: x->  x, y->  z, z-> -y
  Fy: x-> -z, y->  y, z->  x
  Fz: x->  y, y-> -x, z->  z

  il est facilement de dmontrer que l'opration compose est communitative pour ces functions,
  (une rotation en x, puis une en y et quivalent en une en y, puis en x)
  nous noterons l'ensemble des rotations par la compose
  Fx,i o Fy,j o Fz,k avec (i, j, k) dans [0..3]  

*)

  TPiece = class(TObject)
  private
    FName: Char;
    FBaseCubeList: TCubeList;
    FAngleViewList: TList;    //de TCubeList
    FUsed: boolean;
    FColor: TColor;
  public
    constructor Create(_Name: Char; _Color: TColor);
    destructor Destroy; override;

    procedure ComputeAngleViews;
    function AngleViewIsInList(aCoords: TCubeList): boolean;

    function GetBaseCubeString: string;
    function GetAngleViewString(ix: integer): string;

    property Name: Char read FName;
    property BaseCubeList: TCubeList read FBaseCubeList write FBaseCubeList;
    property AngleViewList: TList read FAngleViewList write FAngleViewList;
    property Used: boolean read FUsed write FUsed;
    property Color: TColor read FColor write FColor;
  end;

  TPieceSet = class(TList)
  private
    function TotalCubes: integer;
  public
    procedure Clear; override;
    property CubesCount: integer read TotalCubes;
  end;


  procedure Fx(var _Cube: TCube); overload;
  procedure Fx(var _Cube: TCube; i: integer); overload; //i = la compos

  procedure Fy(var _Cube: TCube); overload;
  procedure Fy(var _Cube: TCube; i: integer); overload; //i = la compos

  procedure Fz(var _Cube: TCube); overload;
  procedure Fz(var _Cube: TCube; i: integer); overload; //i = la compos

implementation

//----------------------------------------------------------------------------------------------------------------------
//
//  Les rotations
//
//----------------------------------------------------------------------------------------------------------------------
procedure Fx(var _Cube: TCube);
var
  _y: integer;
  _z: integer;
begin
  _y:=   _Cube.Z;
  _z:= - _Cube.Y;

  _Cube.Y:= _y;
  _Cube.Z:= _z;
end;

//----------------------------------------------------------------------------------------------------------------------
// Fx o Fx o Fx, i fois, i=0 => pas de rotation donc
procedure Fx(var _Cube: TCube; i: integer);
var
  j: integer;
begin
  for j:= 0 to i - 1 do
    Fx(_Cube);
end;

//----------------------------------------------------------------------------------------------------------------------
procedure Fy(var _Cube: TCube);
var
  _x: integer;
  _z: integer;
begin
  _x:= - _Cube.Z;
  _z:=   _Cube.X;

  _Cube.X:= _x;
  _Cube.Z:= _z;
end;

//----------------------------------------------------------------------------------------------------------------------
procedure Fy(var _Cube: TCube; i: integer);
var
  j: integer;
begin
  for j:= 0 to i - 1 do
    Fy(_Cube);
end;

//----------------------------------------------------------------------------------------------------------------------
procedure Fz(var _Cube: TCube);
var
  _x: integer;
  _y: integer;
begin
  _x:=   _Cube.Y;
  _y:= - _Cube.X;

  _Cube.X:= _x;
  _Cube.Y:= _y;
end;

//----------------------------------------------------------------------------------------------------------------------
procedure Fz(var _Cube: TCube; i: integer);
var
  j: integer;
begin
  for j:= 0 to i - 1 do
    Fz(_Cube);
end;

//----------------------------------------------------------------------------------------------------------------------
//    TPiece
//----------------------------------------------------------------------------------------------------------------------
constructor TPiece.Create(_Name: Char; _Color: TColor);
begin
  inherited Create;

  FName:= _Name;
  FBaseCubeList:= TCubeList.Create;
  FAngleViewList:= TList.Create;
  FUsed:= false;
  FColor:= _Color;
end;


//----------------------------------------------------------------------------------------------------------------------
destructor TPiece.Destroy;
var
  i: integer;
begin
  for i:= FAngleViewList.Count - 1 downto 0 do
    TCubeList(FAngleViewList[i]).Free;
  FAngleViewList.Free;

  FBaseCubeList.Free;

  inherited;
end;

//----------------------------------------------------------------------------------------------------------------------
function TPiece.GetBaseCubeString: string;
begin
  result:= Self.FName + ' ' + FBaseCubeList.GetString;
end;

//----------------------------------------------------------------------------------------------------------------------
function TPiece.GetAngleViewString(ix: integer): string;
begin
  result:= Self.FName + ' View ' + IntToStr(ix) + ' ';

  with TCubeList(FAngleViewList[ix]) do
      result:= Result + GetString;
end;

//----------------------------------------------------------------------------------------------------------------------
//  Comparaison des cubes de la liste FAngleViewList avec la liste de cube pass en paramtre
//  Les listes ont t tries par min d'abord, et ramen  (0,0,0), ce qui nous permet de ne comparer
//  que les cubes de mme indice
function TPiece.AngleViewIsInList(aCoords: TCubeList): boolean;
var
  i, j: integer;
  _Coords: TCubeList;
  continue: boolean;
begin
  result:= false;
  continue:= false;

  for i:= 0 to FAngleViewList.Count - 1 do       //Pour chaque vue
  begin

    _Coords:= TCubeList(FAngleViewList[i]);     // On compare les cubes de la vue au cubes de la liste en parametre

    for j:= 0 to aCoords.Count - 1 do           //Chaque cube de la vue en cours
    begin
                                                //Ds qu'un cube est diffrent on arrete, cela suffit  invalider l'galit
      if ( TCube(aCoords[j]).X) <> TCube(_Coords[j]).X then
        continue:= false
      else
        if ( TCube(aCoords[j]).Y) <> TCube(_Coords[j]).Y then
          continue:= false
        else
          if ( TCube(aCoords[j]).Z) <> TCube(_Coords[j]).Z then
            continue:= false
          else
            continue:= true;

      if not continue then Break;
    end;

    result:= continue;

    if result then break;

  end;

end;

//----------------------------------------------------------------------------------------------------------------------
//  Calcul de l'ensemble (diminu) des angles de vues de la pice.
//  A chaque angle de vue calcule, on tri les cubes et on ramene le cube minimum  (0,0,0)
//  de faon  liminer les angles de vues similaires (vue identique mais
//  coordonnes diffrentes, dcalage d au rotation c'est pour cela qu'on se ramne toujours  (0,0,0) )
//
//  Pour tout i, j, k dans [0..3] on calcule  Fx(i) o Fy(j) o Fz(k)
//
procedure TPiece.ComputeAngleViews;
var
  i, j, k: integer;
  cnt: integer;
  aCoords: TCubeList;
  aCube: TCube;
begin
  FAngleViewList.Clear;

  for i:= 0 to 3 do
    for j:= 0 to 3 do
      for k:= 0 to 3 do
      begin
        aCoords:= TCubeList.Create;

        //Calcul de l'angle de vue
        for cnt:= 0 to FBaseCubeList.Count - 1 do
        begin
          with TCube(FBaseCubeList[cnt]) do
            aCube:= TCube.Create(X, Y, Z);

            Fx(aCube, i);
            Fy(aCube, j);
            Fz(aCube, k);

            aCoords.Add(aCube);
        end;

        //On ramene le cube minimum de la pice  (0, 0, 0)
        aCoords.MoveMinCoordTo000;

        //On vrifie si la vue existe dans la liste des vues
        if not AngleViewIsInList(aCoords) then
          FAngleViewList.Add(aCoords)
        else
          aCoords.Free;

      end;

end;

//----------------------------------------------------------------------------------------------------------------------
//    TPieceList
//----------------------------------------------------------------------------------------------------------------------
procedure TPieceSet.Clear;
var
  i: integer;
begin
  for i:= Self.Count - 1 downto 0 do
    TPiece(Self.Items[i]).Free;

  inherited;
end;

//----------------------------------------------------------------------------------------------------------------------
// Totalise le nombre de cube du set de pices
function TPieceSet.TotalCubes: integer;
var
  i: integer;
begin
  result:= 0;
  for i:= 0 to Self.Count - 1 do
    result:= result + TPiece(Self.Items[i]).FBaseCubeList.Count;
end;


end.


